import re

from analysis.shared.matrix import Matrix
from analysis.annotate.report.annotate_tsv import AnnotateTsv
from analysis.shared.util.labels import letter_label_to_int_label, \
    int_label_to_letter_label
from .slice_matrix import SliceMatrix

NAMES_REGEX = '[A-Z]+(-[A-Z]+)?(\\,[ ]?[A-Z]+(-[A-Z]+)?)*'
CLUSTER_REGEX = 'cluster\\([0-9]+\\)'


class SliceCluster(object):
    def __init__(self, matrix, annotate_tsv, clustersnames):
        self.matrix = matrix
        self.annotate_tsv = annotate_tsv
        self.clustersnames = clustersnames

    def parse_cluster_names(self, raw_names):
        """
        names raw list of names
        return list of cluster names
        raise Exception
        """
        names_str = ' '.join(raw_names)

        if not re.match(NAMES_REGEX, names_str):
            raise Exception(
                'Bad format for the clusters groups. Must be something like \
                    that: "A-C, E"')

        names_subgroups = [sbg.strip() for sbg in names_str.split(',')]

        result = []
        for subgroup in names_subgroups:
            result.extend(
                self.convert_relative_clusters_to_explicit_list(subgroup))

        if len(result) != len(set(result)):
            raise Exception('Some clusters groups are prensent multiple times')

        return result

    def convert_relative_clusters_to_explicit_list(self, subcluster):
        indvidual_letters = subcluster.split('-')

        if len(indvidual_letters) == 1:
            return indvidual_letters
        elif len(indvidual_letters) == 2:
            return self.make_letter_range(*indvidual_letters)
        else:
            raise Exception(
                'Bad format for the clusters groups catched too late')

    def make_letter_range(self, first_s, second_s):
        first = letter_label_to_int_label(first_s)
        second = letter_label_to_int_label(second_s)

        if second <= first:
            raise Exception('Bad letters range: ' + first_s + '-' + second_s)

        return [int_label_to_letter_label(x) for x in range(first, second + 1)]

    def extract_wanted_file_names(self, annotate_tsv, names):
        cluster_header = self.find_cluster_header(annotate_tsv)
        row_indexes = annotate_tsv.make_filter(
            cluster_header, lambda x: x in names)
        return annotate_tsv.list_items(
            annotate_tsv.unique_identifier, row_indexes)

    def find_cluster_header(self, annotate_tsv):
        regex = re.compile(CLUSTER_REGEX)

        for header in annotate_tsv.list_headers():
            matched = regex.match(header)

            if matched:
                return header

        raise Exception('The annotate tsv file contains no cluster')

    def run(self):
        matrix = Matrix.parse_matrixfile(self.matrix)
        annotate_tsv = AnnotateTsv.load_annotate_tsv(self.annotate_tsv)
        names = self.parse_cluster_names(self.clustersnames)
        wanted_file_names = self.extract_wanted_file_names(annotate_tsv, names)

        print(SliceMatrix.slice(matrix, wanted_file_names), end='')



from .launcher import SliceFilenameLauncher, SliceClusterLauncher


class SliceFilenameParser(object):
    @staticmethod
    def add_parser(subparsers):
        parser = subparsers.add_parser(
            'slice-filename',
            description='Slice the matrix to keep specified filename(s).',
            help='Slice the matrix to keep specified filename(s).'
        )
        parser.set_defaults(func=SliceFilenameLauncher.run)

        parser.add_argument(
            'matrix', help='The input matrix tsv file.', type=str)
        parser.add_argument(
            'file_names', help='The input file_names text file.', type=str)


class SliceClusterParser(object):
    @staticmethod
    def add_parser(subparsers):
        parser = subparsers.add_parser(
            'slice-cluster',
            description='Slice the matrix to keep specified cluster(s).',
            help='Slice the matrix to keep specified cluster(s).'
        )
        parser.set_defaults(func=SliceClusterLauncher.run)

        parser.add_argument(
            'matrix', help='The input matrix tsv file.', type=str)
        parser.add_argument(
            'annotate_tsv', help='The input annotate tsv file.', type=str)
        parser.add_argument(
            'clustersnames', help='The clusters names (e.g. "A-E, H").',
            type=str)



from analysis.ari.breakdowns import Breakdowns
from analysis.shared.metadata import Metadata
from analysis.shared.matrix import Matrix
import warnings
warnings.filterwarnings("ignore")


DESCRIPTION = '{matrix_description}, clusteringAlgorithm={clustering_algorithm}\
    , linkageMethod={linkage_method}'


class AriApp:
    def __init__(self, params):
        self.matrix_filename = params['matrix_filename']
        self.metadata_filename = params['metadata_filename']
        self.linkage_method = params['linkage_method']
        self.breakdown_categories = params['breakdown_categories']
        self.restriction_categories = params['restriction_categories']
        self.title = params['title']
        self.ignore_user = params['ignore_user']
        self.md5sum = params['md5sum']

    def run(self):
        # load files
        matrix = self._load_matrix()
        metadata = self._load_metadata()

        if self.ignore_user:
            matrix = matrix.sub_matrix(metadata.filenames)

        # parse categories
        restrictions = ['md5sum', 'file_path',
                        'file_name', 'id', 'virtual', 'assembly']
        ordered_categories = ['assay', 'publishing_group', 'cell_type']

        breakdown_categories = metadata.make_usable_categories(
            matrix.get_file_names(),
            self.breakdown_categories,
            restrictions,
            ordered_categories)
        restriction_categories = metadata.make_usable_categories(
            matrix.get_file_names(),
            self.restriction_categories,
            restrictions,
            ordered_categories)

        # Make ARI Annotation
        breakdowns = Breakdowns.make_breakdowns(matrix,
                                                metadata,
                                                breakdown_categories,
                                                restriction_categories,
                                                self.linkage_method)

        # return report
        return self.breakdowns_to_tsv(breakdowns, matrix.description)

    def breakdowns_to_tsv(self, breakdowns, matrix_description):
        return '\n'.join([self.make_first_line(matrix_description),
                          self.make_ari_lines(breakdowns),
                          self.make_seperator(),
                          self.make_nb_tags_lines(breakdowns)])

    def make_first_line(self, matrix_description):
        return '\t'.join(
            [
                self.title,
                DESCRIPTION.format(
                    matrix_description=matrix_description,
                    clustering_algorithm='hierarchical',
                    linkage_method=self.linkage_method)])

    def make_ari_lines(self, breakdowns):
        header = '\t'.join(['ARI_per_tags_in_category'] +
                           breakdowns.get_restriction_categories() +
                           ['total_nb_datasets'])
        return '\n'.join([header] +
                         [self.make_ari_line(breakdown)
                          for breakdown in breakdowns])

    def make_ari_line(self, breakdown):
        return '\t'.join([breakdown.get_name()] +
                         ['{:.3f}'.format(ari)
                          for ari in breakdown.get_aris()] +
                         [str(breakdown.get_dataset_size())])

    def make_seperator(self):
        return ''

    def make_nb_tags_lines(self, breakdowns):
        header = '\t'.join(['nb_tags_per_category'] +
                           breakdowns.get_restriction_categories() +
                           ['total_nb_datasets'])
        return '\n'.join([header] +
                         [self.make_nb_tags_line(breakdown)
                          for breakdown in breakdowns])

    def make_nb_tags_line(self, breakdown):
        return '\t'.join([breakdown.get_name()] +
                         [str(nb_clusters)
                          for nb_clusters in breakdown.get_nbs_clusters()] +
                         [str(breakdown.get_dataset_size())])

    def _load_matrix(self):
        return Matrix.parse_matrixfile(self.matrix_filename)

    def _load_metadata(self):
        if self.md5sum:
            unique_identifier = 'md5sum'
        else:
            unique_identifier = 'file_name'

        return Metadata.parse_metadatafile(
            self.metadata_filename, unique_identifier)



from analysis.annotate.annotateapp import AnnotateApp
from analysis.shared.exception.geecanalysisexception import \
    GeecAnalysisException


class AnnotateAppBuilderException(GeecAnalysisException):
    pass


class ParameterAnnotateAppBuilderException(AnnotateAppBuilderException):
    pass


class AnnotateAppBuilder:
    def __init__(self):
        self.params = {
            'matrix_filename': None,
            'metadata_filename': None,
            'groups': None,
            'title': None,
            'linkage_method': None,
            'desired_nb_clusters': None,
            'pdf_file': None,
            'tsv_file': None,
            'reordered_matrix_filename': None,
            'seed': None,
            'mds': False,
            'umap': False,
            'rescale': False,
            'md5sum': False
        }
        self.needed = ['matrix_filename', 'metadata_filename',
                       'groups', 'title', 'linkage_method']

    def set_title(self, title):
        self.params['title'] = title

    def set_matrix_filename(self, matrix_filename):
        self.params['matrix_filename'] = matrix_filename

    def set_metadata_filename(self, metadata_filename):
        self.params['metadata_filename'] = metadata_filename

    def set_desired_nb_clusters(self, desired_nb_clusters):
        self.params['desired_nb_clusters'] = desired_nb_clusters

    def set_groups(self, groups):
        self.params['groups'] = groups

    def set_pdf_file(self, pdf_file):
        self.params['pdf_file'] = pdf_file

    def set_tsv_file(self, tsv_file):
        self.params['tsv_file'] = tsv_file

    def set_linkage_method(self, linkage_method):
        self.params['linkage_method'] = linkage_method

    def set_reordered_matrix_filename(self, reordered_matrix_filename):
        self.params['reordered_matrix_filename'] = reordered_matrix_filename

    def set_seed(self, seed):
        self.params['seed'] = seed

    def set_mds(self, mds):
        self.params['mds'] = mds

    def set_umap(self, umap):
        self.params['umap'] = umap

    def set_rescale(self, rescale):
        self.params['rescale'] = rescale

    def set_md5sum(self, md5sum):
        self.params['md5sum'] = md5sum

    def build(self):
        for p in self.needed:
            if self.params.get(p) is None:
                raise ParameterAnnotateAppBuilderException(
                    'The parameter:\'' + p + '\' is not specified.')

        return AnnotateApp(self.params)



"""Dimentional reduction algorithms
"""
import sklearn.manifold as sm
import umap


class MDSManifoldFactory:
    """A Multidimensional Scaling Manifold factory

    Parameters
    ----------
    random_state : int, RandomState instance, default=None
        Determines the random number generator used to initialize the centers.
        Pass an int for reproducible results across multiple function calls.

    See Also
    --------
    sklearn.manifold.MDS
    """

    def __init__(self, random_state=None):
        self.random_state = random_state

    def make(self):
        """Make the MDS manifold

        Returns
        -------
        sklearn.manifold.MDS
            return the MDS manifold object

        See Also
        --------
        sklearn.manifold.MDS
        """
        return sm.MDS(dissimilarity='precomputed', n_jobs=-1,
                      random_state=self.random_state)


class UMAPManifoldFactory:
    """A UMAP Manifold factory

    Parameters
    ----------
    random_state : int, RandomState instance, default=None
        Determines the random number generator used to initialize the centers.
        Pass an int for reproducible results across multiple function calls.

    See Also
    --------
    umap.UMAP
    """

    def __init__(self, random_state=None):
        self.random_state = random_state

    def make(self):
        """Make the UMAP manifold

        Returns
        -------
        umap.UMAP
            return the UMAP manifold object

        See Also
        --------
        umap.UMAP
        """
        return umap.UMAP(random_state=self.random_state)



import pylab
import matplotlib.pyplot as plt
import matplotlib
from .composer import Composer, ScatterPlot
import warnings
warnings.filterwarnings("ignore")
matplotlib.use('Agg')


class MdsPageComposer(Composer):
    def __init__(self, p_matrix, annotation, group, title, method):
        self.p_matrix = p_matrix
        self.annotation = annotation
        self.group = group
        self.title = title
        self.method = method

    def run(self, left_lower_page, right_lower_page):
        y_size = 1080*2
        x_size = (14 * y_size) / 8.5

        fig = pylab.figure(figsize=(x_size/100., y_size/100.), dpi=100.)

        axtitle = fig.add_axes([0.1, 0.9, 0.8, 0.01])
        axtitle.axis('off')
        pylab.title(self.title, fontsize=56, loc='left')
        axtitle.set_xticks([])
        axtitle.set_yticks([])

        axtitle = fig.add_axes([0.1, 0.870, 0.8, 0.01])
        axtitle.axis('off')
        pylab.title("Method: " + self.method, fontsize=32, loc='left')
        axtitle.set_xticks([])
        axtitle.set_yticks([])

        axtitle = fig.add_axes([0.1, 0.840, 0.8, 0.01])
        axtitle.axis('off')
        pylab.title("Attribute: " + self.group, fontsize=32, loc='left')
        axtitle.set_xticks([])
        axtitle.set_yticks([])

        self._group_scatter_plot(fig)
        self._cluster_scatter_plot(fig)

        # time and page
        axid = fig.add_axes([0., 0., 1., 0.015])
        axid.axis('off')
        axid.text(0.005, 0.25, left_lower_page)
        axid.text(0.995, 0.25, right_lower_page, ha='right')
        axid.set_xticks([])
        axid.set_yticks([])

        filename = self.tmp_name()
        plt.savefig(filename, format='png')
        return filename

    def _draw_scatter_plot(self, fig, axes, column, scatter_plot):
        p_matrix = scatter_plot.get_p_matrix()
        ax = fig.add_axes(axes)
        ax.scatter(p_matrix[:, 0],
                   p_matrix[:, 1],
                   s=100,
                   cmap='rgb',
                   c=scatter_plot.get_colors())
        pylab.title(scatter_plot.get_title(),
                    fontsize='xx-large')
        plt.legend(handles=scatter_plot.get_handles(),
                   bbox_to_anchor=(0., -0.03),
                   loc=2,
                   ncol=column)

    def _group_scatter_plot(self, fig):
        self._draw_scatter_plot(fig,
                                [0.08, 0.15, .4, .6588235294],
                                3,
                                ScatterPlot.make_tags_sp(self.group,
                                                         self.p_matrix,
                                                         self.annotation,
                                                         15))

    def _cluster_scatter_plot(self, fig):
        self._draw_scatter_plot(fig,
                                [0.52, 0.15, .4, .6588235294],
                                11,
                                ScatterPlot.make_clusters_sp(self.p_matrix,
                                                             self.annotation))



from .ariapp_builder import AriAppBuilder


class EvaluateLauncher(object):
    @staticmethod
    def run(args):
        ab = AriAppBuilder()
        ab.set_matrix_filename(args.matrix)
        ab.set_metadata_filename(args.metadata)
        ab.set_linkage_method(args.linkage)
        ab.set_breakdown_categories(args.b)
        ab.set_restriction_categories(args.r)
        ab.set_title(args.title)
        ab.set_ignore_user(args.i)
        ab.set_md5sum(args.md5sum)

        print(ab.build().run())



from .input_parser import InputParser
from .shared.exception.geecanalysisexception import GeecAnalysisException
from .shared.exception.geecanalysiserror import GeecAnalysisError
import logging
import os
import os.path
import sys
sys.path.append(os.path.dirname(os.path.realpath(__file__)))


sys.setrecursionlimit(10000)
logging.basicConfig()
logger = logging.getLogger(__name__)


def main(argv):
    try:
        args = InputParser.parse_args(argv)
        args.func(args)
    except GeecAnalysisError as e:
        logger.error(e)
    except GeecAnalysisException as e:
        logger.critical(e, exc_info=e)
    except Exception as e:
        logger.exception(e)


def cli():
    main(sys.argv[1:])


if __name__ == '__main__':
    cli()



import re


class StringBuilder():
    def __init__(self, template):
        self.template = template
        self.params = {}

    def build(self):
        tags = re.findall('%\\((.*?)\\)', self.template)
        template_tags = [(tag, self.params.get(tag, 'None')) for tag in tags]

        transformed = self.template
        for tag, value in template_tags:
            transformed = re.sub('%\\((' + tag + ')\\)',
                                 str(value), transformed)

        return transformed

    def set(self, key, value):
        self.params[key] = value



import os
import tempfile
from itertools import cycle
import matplotlib
import matplotlib.patches as mpatches
import numpy

OTHERS = 'Others'
USER = 'User'


class Composer:
    def run(self, filename, left_lower_page, right_lower_page):
        raise NotImplementedError('Composer.run() is an abstract method')

    def tmp_name(self):
        fd, temp_path = tempfile.mkstemp()
        os.close(fd)
        os.remove(temp_path)
        return temp_path


class LegendMaker:
    KELLY_COLORS = [(1.0, 0.7019607843137254, 0.0),
                    (0.5019607843137255, 0.24313725490196078,
                     0.4588235294117647),
                    (1.0, 0.40784313725490196, 0.0),
                    (0.6509803921568628, 0.7411764705882353,
                     0.8431372549019608),
                    (0.7568627450980392, 0.0, 0.12549019607843137),
                    (0.807843137254902, 0.6352941176470588,
                     0.3843137254901961),
                    (0.5058823529411764, 0.4392156862745098, 0.4),
                    (0.0, 0.49019607843137253, 0.20392156862745098),
                    (0.9647058823529412, 0.4627450980392157,
                     0.5568627450980392),
                    (0.0, 0.3254901960784314, 0.5411764705882353),
                    (1.0, 0.47843137254901963, 0.3607843137254902),
                    (0.3254901960784314, 0.21568627450980393,
                     0.47843137254901963),
                    (1.0, 0.5568627450980392, 0.0),
                    (0.7019607843137254, 0.1568627450980392,
                     0.3176470588235294),
                    (0.9568627450980393, 0.7843137254901961, 0.0),
                    (0.4980392156862745, 0.09411764705882353,
                     0.050980392156862744),
                    (0.5764705882352941, 0.6666666666666666, 0.0),
                    (0.34901960784313724, 0.2, 0.08235294117647059),
                    (0.9450980392156862, 0.22745098039215686,
                     0.07450980392156863),
                    (0.13725490196078433, 0.17254901960784313,
                     0.08627450980392157)]
    USER_COLOR = (0., 0.85, 0.85)

    @staticmethod
    def create_colors_dict(labels):
        rri = cycle(LegendMaker.KELLY_COLORS)
        colors = []

        for lbl in labels:
            if lbl == USER:
                colors.append(
                    (lbl, matplotlib.colors.rgb2hex(LegendMaker.USER_COLOR)))
            else:
                colors.append((lbl, matplotlib.colors.rgb2hex(next(rri))))

        return dict(colors)

    @staticmethod
    def create_patches(labels, colors_dict, formatted_labels):
        patches = []

        if USER in labels:
            patches.append(mpatches.Patch(
                color=colors_dict[USER], label=formatted_labels[USER]))

        patches += [mpatches.Patch(color=colors_dict[lbl],
                                   label=formatted_labels[lbl])
                    for lbl in labels
                    if lbl != USER]

        return patches

    @staticmethod
    def limit_chars(string, max_chars):
        if len(string) > max_chars:
            return string[:max_chars - 1] + u'\0x2026'
        else:
            return string

    @staticmethod
    def make_composed_tags(tag_names, max_nb_tags):
        firsts = tag_names[:max_nb_tags - 1]
        lasts = tag_names[max_nb_tags - 1:]

        if USER in lasts:
            lasts.remove(USER)
            lasts.insert(0, firsts.pop())
            firsts.insert(0, USER)

        composed_tags = [(f, [f]) for f in firsts]

        if len(lasts) > 1:
            composed_tags += [('Others', lasts)]
        else:
            composed_tags += [(last, [last]) for last in lasts]

        return composed_tags


class ScatterPlot:
    def __init__(self, title, p_matrix, colors, handles):
        self.title = title
        self.handles = handles
        self.p_matrix = p_matrix
        self.colors = colors

    def get_title(self):
        return self.title

    def get_handles(self):
        return self.handles

    def get_p_matrix(self):
        return self.p_matrix

    def get_colors(self):
        return self.colors

    @staticmethod
    def make_tags_sp(category_name, p_matrix, annotation, max_nb_tags):
        tag_names_with_len = annotation.obtain_size_of_tags_in_category(
            category_name)
        tag_names_with_len.sort(key=lambda x: x[1], reverse=True)

        tag_names = [t for (t, _) in tag_names_with_len]

        composed_tags = LegendMaker.make_composed_tags(tag_names, max_nb_tags)
        labels = [c[0] for c in composed_tags]

        formatted_labels = {lbl: LegendMaker.limit_chars(
            lbl, 20) for lbl in labels}
        colors_dict = LegendMaker.create_colors_dict(labels)

        handles = LegendMaker.create_patches(
            labels, colors_dict, formatted_labels)

        composed_tags_dict = {tag: ct[0]
                              for ct in composed_tags for tag in ct[1]}
        ordered_tag_names = annotation.obtain_ordered_tag_names(category_name)

        colors = [colors_dict[composed_tags_dict[tag_name]]
                  for tag_name in ordered_tag_names]

        non_user_ids = [index
                        for index, tag_name in enumerate(ordered_tag_names)
                        if composed_tags_dict[tag_name] != USER]
        user_ids = [index
                    for index, tag_name in enumerate(ordered_tag_names)
                    if composed_tags_dict[tag_name] == USER]

        reordered_p_matrix = numpy.concatenate(
            (p_matrix[non_user_ids], p_matrix[user_ids]))
        reordered_colors = [colors[idx] for idx in non_user_ids + user_ids]

        return ScatterPlot('Groups', reordered_p_matrix, reordered_colors,
                           handles)

    @staticmethod
    def make_clusters_sp(p_matrix, annotation):
        cluster_ids = range(1, annotation.get_nb_clusters() + 1)

        formatted_labels = {
            cluster_id: annotation.obtain_cluster_name(cluster_id)
            for cluster_id in cluster_ids}
        colors_dict = LegendMaker.create_colors_dict(cluster_ids)
        handles = LegendMaker.create_patches(
            cluster_ids, colors_dict, formatted_labels)

        colors = [colors_dict[cluster_id]
                  for cluster_id in annotation.make_flat_clusters()]

        return ScatterPlot('Clusters', p_matrix, colors, handles)


class PieChartsBuilder:
    def __init__(self):
        self.annotation = None
        self.category_name = None
        self.max_nb_tags = 0x7FFFFFFFFFFFFFFF  # int64 max
        self.show_nb_tags = False
        self.max_chars = -1

    def build(self):
        if self.annotation is None or self.category_name is None:
            raise Exception('No annotation or no category_name passed')

        tag_names_with_len = self.annotation.obtain_size_of_tags_in_category(
            self.category_name)
        tag_names_with_len.sort(key=lambda x: x[1], reverse=True)

        tag_names = [t for (t, _) in tag_names_with_len]

        composed_tags = LegendMaker.make_composed_tags(
            tag_names, self.max_nb_tags)
        sizes = self._list_size_for_composed_tags(
            composed_tags, tag_names_with_len)
        formatted_labels = self._make_labels_for_tags(composed_tags, sizes)

        return PieCharts.make_piecharts(self.annotation, self.category_name,
                                        composed_tags, formatted_labels)

    def _make_labels_for_tags(self, composed_tags, sizes):
        if self.show_nb_tags:
            labels = map(lambda x, y: '(' + str(x) + ') ' +
                         y[0], sizes, composed_tags)
        else:
            labels = [ct[0] for ct in composed_tags]

        if self.max_chars > 0:
            labels = [LegendMaker.limit_chars(
                lbl, self.max_chars) for lbl in labels]

        return {ct[0]: lbl for ct, lbl in zip(composed_tags, labels)}

    def _list_size_for_composed_tags(self, composed_tags, tag_names_with_len):
        lenghts = {x: y for x, y in tag_names_with_len}
        return [sum([lenghts[lbl]
                     for lbl in ct[1]])
                for ct in composed_tags]

    def set_annotation(self, annotation):
        self.annotation = annotation

    def set_category_name(self, category_name):
        self.category_name = category_name

    def set_max_nb_tags(self, max_nb_tags):
        self.max_nb_tags = max_nb_tags

    def set_show_nb_tags(self, show_nb_tags):
        self.show_nb_tags = show_nb_tags

    def set_max_chars(self, max_chars):
        self.max_chars = max_chars


class PieCharts(dict):
    def __init__(self, category_name, piecharts, handles):
        super(PieCharts, self).__init__(piecharts)
        self.category_name = category_name
        self.handles = handles

    def get_percents_for_cluster(self, cluster_id):
        return self[cluster_id].get_percents()

    def get_colors_for_cluster(self, cluster_id):
        return self[cluster_id].get_colors()

    def get_category_name(self):
        return self.category_name

    def get_handles(self):
        return self.handles

    def get_title_for_cluster(self, cluster_id):
        return self[cluster_id].get_title()

    @staticmethod
    def make_piecharts(annotation, category_name, composed_tags,
                       formatted_labels):
        # Make dict of translation
        # create patch and color dict
        labels = [c[0] for c in composed_tags]
        colors_dict = LegendMaker.create_colors_dict(labels)
        handles = LegendMaker.create_patches(
            labels, colors_dict, formatted_labels)

        return PieCharts(category_name,
                         {
                             cluster_id: PieChart.make_piechart(
                                 annotation, cluster_id, category_name,
                                 composed_tags, colors_dict)
                             for cluster_id
                             in range(annotation.get_nb_clusters() + 1)},
                         handles)


class PieChart:
    def __init__(self, name, slices):
        self.name = name
        self.slices = slices

    def get_percents(self):
        return self.slices.get_percents()

    def get_title(self):
        return self.name + '=' + str(self.slices.get_total())

    def get_colors(self):
        return self.slices.get_colors()

    @staticmethod
    def make_piechart(annotation, cluster_id, category_name, composed_tags,
                      colors_dict):
        return PieChart(annotation.obtain_cluster_name(cluster_id),
                        Slices.make_slices(annotation, cluster_id,
                                           category_name, composed_tags,
                                           colors_dict))


class Slices(list):
    def __init__(self, *args):
        super(Slices, self).__init__(*args)

    def get_percents(self):
        counts = [float(x.get_count()) for x in self]
        total = float(sum(counts))
        return [count/total for count in counts]

    def get_total(self):
        return sum([x.get_count() for x in self])

    def get_colors(self):
        return [x.get_color() for x in self]

    @staticmethod
    def make_slices(annotation, cluster_id, category_name, composed_tags,
                    colors_dict):
        return Slices([Slice.make_slice(annotation, cluster_id, category_name,
                                        ct, colors_dict)
                       for ct in composed_tags
                       if annotation.obtain_size_of_tags_from_tag_names(
                           cluster_id, category_name, ct[1]) > 0])


class Slice:
    def __init__(self, name, count, color):
        self.name = name
        self.color = color
        self.count = count

    def get_name(self):
        return self.name

    def get_count(self):
        return self.count

    def get_color(self):
        return self.color

    @staticmethod
    def make_slice(annotation, cluster_id, category_name, composed_tag,
                   colors_dict):
        return Slice(composed_tag[0],
                     annotation.obtain_size_of_tags_from_tag_names(
                         cluster_id, category_name, composed_tag[1]),
                     colors_dict[composed_tag[0]])



from .cat import run as run_cat
from .grep import run as run_grep


def select_unique_identifier(arg):
    if arg:
        unique_identifier = 'md5sum'
    else:
        unique_identifier = 'file_name'
    return unique_identifier


class CatLaucher(object):
    @staticmethod
    def run(args):
        run_cat(args.input, select_unique_identifier(args.md5sum))


class GrepLauncher(object):
    @staticmethod
    def run(args):
        run_grep(
            args.inputs,
            args.pattern,
            args.e,
            args.f,
            args.v,
            select_unique_identifier(args.md5sum)
        )



from analysis.shared.util.labels import int_label_to_letter_label


class Clusters(dict):
    def __init__(self, clusters_dict, linkage_matrix, linkage_method,
                 clustering_algorithm):
        super(Clusters, self).__init__(clusters_dict)
        self.clusters_ref = {
            file_name: cl
            for cl in self.values()
            for file_name in cl.get_file_names()}
        self._linkage_matrix = linkage_matrix
        self._linkage_method = linkage_method
        self._clustering_algorithm = clustering_algorithm

    def get_file_names(self):
        return self.clusters_ref.keys()

    def obtain_cluster_name(self, file_name):
        return self.clusters_ref[file_name].get_name()

    def make_flat_clusters(self, matrix):
        return [self.clusters_ref[fn].get_id()
                for fn in matrix.get_file_names()]

    @property
    def linkage_matrix(self):
        return self._linkage_matrix

    @property
    def linkage_method(self):
        return self._linkage_method

    @property
    def clustering_algorithm(self):
        return self._clustering_algorithm

    @property
    def count(self):
        return len(self)

    @staticmethod
    def make_clusters(cluster_ids, file_names, linkage_matrix, linkage_method,
                      clustering_algorithm):
        if len(cluster_ids) != len(file_names):
            raise Exception('The length of the two list aren\'t matching!')

        clusters_file_names = {cluster_id: []
                               for cluster_id in set(cluster_ids)}

        for cluster_id, file_name in zip(cluster_ids, file_names):
            clusters_file_names[cluster_id].append(file_name)

        return Clusters({cls_id: Cluster.make_cluster(cls_id, md5s)
                         for cls_id, md5s in clusters_file_names.items()},
                        linkage_matrix,
                        linkage_method,
                        clustering_algorithm)

    def remap_clusters_ids(self, new_clusters_ids, matrix):
        return Clusters.make_clusters(
            new_clusters_ids, matrix.get_file_names(), self.linkage_matrix,
            self.linkage_method, self.clustering_algorithm)


class Cluster:
    def __init__(self, cluster_id, cluster_name, file_names):
        self.cluster_id = cluster_id
        self.name = cluster_name
        self.file_names = file_names

    def get_id(self):
        return self.cluster_id

    def get_name(self):
        return self.name

    def get_file_names(self):
        return self.file_names

    def __len__(self):
        return len(self.file_names)

    @staticmethod
    def make_cluster(cluster_id, file_names):
        return Cluster(cluster_id,
                       int_label_to_letter_label(cluster_id),
                       file_names)



import scipy.cluster.hierarchy as sch


class ScipyPatch(object):
    """
    The scipy dendrogram imposes an order for the matrix and the clusters.
    ScipyPatch fixes issues caused with incompatible naming and ordering.
    """

    @staticmethod
    def reorder_matrix(clusters, matrix):
        """
        This method reorder the matrix with the dendogram.
        """
        file_names = matrix.get_file_names()
        Z = ScipyPatch.make_dendrogram(clusters, matrix)
        return matrix.sub_matrix(
            reversed([file_names[leaf] for leaf in Z['leaves']]))

    @staticmethod
    def rename_clusters(clusters, matrix):
        Z = ScipyPatch.make_dendrogram(clusters, matrix)

        real_cluster_order = ScipyPatch._obtain_order(reversed(Z['ivl']))
        wanted_cluster_order = range(1, clusters.count + 1)  # [1..n]

        remap_dict = {r: w for r, w in zip(
            real_cluster_order, wanted_cluster_order)}

        return clusters.remap_clusters_ids(
            [remap_dict[x]
             for x in clusters.make_flat_clusters(matrix)], matrix)

    @staticmethod
    def make_dendrogram(clusters, matrix):
        """
        This method uniformize the way the dendrogram is made with this
        objects methods and the code.
        """
        return sch.dendrogram(
            clusters.linkage_matrix,
            color_threshold=ScipyPatch._compute_color_threshold(
                clusters),
            orientation='left',
            distance_sort='descending',
            labels=clusters.make_flat_clusters(matrix))

    @staticmethod
    def _compute_color_threshold(clusters):
        """
        Compute the color threshold for the dendrogram.

        Reference
        -----
        See color_threshold of scipy.cluster.hierarchy.dendrogram
        """
        return clusters.linkage_matrix[-(clusters.count-1), 2]

    @staticmethod
    def _obtain_order(flat_cluster):
        """
        This function takes a flat_cluster and return the order of the ids.
        """
        inside = set()

        def add_if_not_inside(x):
            if x in inside:
                return False
            else:
                inside.add(x)
                return True

        return filter(add_if_not_inside, flat_cluster)



class SliceMatrix(object):
    @staticmethod
    def slice(matrix, wanted_file_names):
        """
        matrix The Matrix to slice
        wanted_datasets The list of datasets name to keep
        return Sliced Matrix
        """
        return matrix.sub_matrix(
            filter(
                lambda filename: filename in matrix.get_file_names(),
                wanted_file_names
            )
        )



from analysis.shared.matrix import Matrix
from analysis.slice.slice_matrix import SliceMatrix


class SliceFilename(object):
    def __init__(self, matrix, file_names):
        self.matrix = matrix
        self.file_names = file_names

    def geec_slice_file_name(self, matrix_filename, file_names, output=None):
        matrix = Matrix.parse_matrixfile(matrix_filename)
        newmatrix = SliceMatrix.slice(matrix, file_names)

        if output:
            with open(output, 'w') as output:
                output.write(str(newmatrix))
        else:
            print(newmatrix, end='')

    def run(self):
        with open(self.file_names, 'r') as f:
            filenames = f.read().split('\n')

        self.geec_slice_file_name(self.matrix, filenames)



import json
import re
from analysis.shared.exception.geecanalysiserror import GeecAnalysisError
from analysis.shared.ihecjson import IhecJson


class MetadataFormatError(GeecAnalysisError):
    pass


class MetadataIOError(GeecAnalysisError):
    pass


class MetadataFactory:
    @staticmethod
    def empty(unique_identifier):
        return Metadata({'datasets': {}}, unique_identifier)


class Metadata:

    def __init__(self, metadata, unique_identifier):
        self.metadata = _Metadata.make_metadata(metadata, unique_identifier)
        self.unique_identifier = unique_identifier

    @staticmethod
    def parse_metadatafile(filename, unique_identifier):
        try:
            with open(filename, 'r') as json_file:
                return Metadata.load_from_file(json_file, unique_identifier)
        except IOError as e:
            raise MetadataIOError(e)

    @staticmethod
    def load_from_file(json_file, unique_identifier):
        try:
            data = json.load(json_file)
        except ValueError:
            raise MetadataFormatError('The metadata file is not a valid JSON.')

        try:
            if 'hub_description' in data:
                data = IhecJson.epigeec_format(data)

            if type(data['datasets']) is list:
                data['datasets'] = {
                    dataset[unique_identifier]: dataset
                    for dataset in data['datasets']}

        except KeyError:
            raise MetadataFormatError(
                'The metadata file format is not supported.')

        return Metadata(data, unique_identifier)

    def __str__(self):
        return self.metadata.to_json()

    def __eq__(self, other):
        if isinstance(other, Metadata):
            return self.metadata == other.metadata
        return False

    def __ne__(self, other):
        if isinstance(other, Metadata):
            return self.metadata != other.metadata
        return False

    def __cmp__(self, other):
        if isinstance(other, Metadata):
            return self.metadata == other.metadata
        return False

    def __len__(self):
        return len(self.metadata['datasets'])

    def __add__(self, other):
        return Metadata(self.metadata + other.metadata, self.unique_identifier)

    @property
    def filenames(self):
        return self.metadata['datasets'].keys()

    def key_intersection(self, collection):
        return set(self.metadata['datasets'].keys()).intersection(collection)

    def find(self, patterns, is_inverted):
        md5lists = [self._find(pattern, is_inverted)
                    for pattern in patterns]

        file_names = self._fusion_md5lists(md5lists, is_inverted)

        return self.extract(file_names)

    def _find(self, pattern, is_inverted):
        return [file_name
                for file_name, dataset in self.metadata['datasets'].items()
                if any([str(data).find(pattern) >= 0
                        for data in dataset.values()]) != is_inverted]

    def match(self, regexs, is_inverted):
        md5lists = [self._match(regex, is_inverted)
                    for regex in regexs]

        file_names = self._fusion_md5lists(md5lists, is_inverted)

        return self.extract(file_names)

    def _match(self, regex, is_inverted):
        compiled_re = re.compile(regex)
        return [file_name
                for file_name, dataset in self.metadata['datasets'].items()
                if any([compiled_re.match(str(data))
                        for data in dataset.values()]) != is_inverted]

    def _fusion_md5lists(self, md5lists, is_inverted):
        if md5lists:
            first = set(md5lists[0])
            others = md5lists[1:]

            if is_inverted:
                file_names = first.intersection(*others)
            else:
                file_names = first.union(*others)
        else:
            file_names = []

        return file_names

    def extract(self, file_names):
        return Metadata(
            self.metadata.extract(file_names),
            self.unique_identifier)

    def make_usable_categories(self, file_names, wanted_categories=None,
                               unwanted_categories=[], ordered_categories=[]):
        all_categories = self.list_categories(file_names)

        filtered_categories = [
            a_cat
            for a_cat
            in all_categories
            if a_cat not in unwanted_categories]

        # if None or empty
        if not wanted_categories:
            return ([o_cat
                     for o_cat
                     in ordered_categories
                     if o_cat in filtered_categories] +
                    [f_cat
                     for f_cat
                     in filtered_categories
                     if f_cat not in ordered_categories])
        else:
            return [w_cat
                    for w_cat
                    in wanted_categories
                    if w_cat in filtered_categories]

    def list_categories(self, file_names):
        categories_name = set()
        datasets = self.metadata['datasets']

        for file_name in file_names:
            if file_name in datasets:
                dataset = datasets[file_name]
                categories_name.update(dataset.keys())

        return list(categories_name)

    def obtain_formated_dataset(self, file_name, categories_names):
        return [self.obtain_dataset_item(file_name, category_name)
                for category_name in categories_names]

    def obtain_dataset_item(self, file_name, category_name):
        return self.metadata['datasets'][file_name][category_name]

    def extract_tag_names_file_names(self, file_names, category_name):
        tags = {}

        for file_name in file_names:
            tag_name = self.obtain_dataset_item(file_name, category_name)

            if tag_name not in tags:
                tags[tag_name] = [file_name]
            else:
                tags[tag_name].append(file_name)

        return tags


class _Metadata(dict):
    def __init__(self, tuples, unique_identifier):
        super(_Metadata, self).__init__(tuples)
        self.unique_identifier = unique_identifier

    def __add__(self, other):
        generic_keys = self._extract_generic_keys(other)
        datasets = self['datasets'] + other['datasets']

        tuples = [('datasets', datasets),
                  ('count', len(datasets))]
        tuples += self._fusion_metadata_values(other, generic_keys)

        return _Metadata(tuples, self.unique_identifier)

    def extract(self, file_names):
        generic_keys = self._extract_generic_keys()
        datasets = self['datasets'].extract(file_names)

        tuples = [('datasets', datasets),
                  ('count', len(datasets))]
        tuples += [(generic_key, self[generic_key])
                   for generic_key in generic_keys]

        return _Metadata(tuples, self.unique_identifier)

    def _extract_generic_keys(self, other={}):
        fusioned_keys = set(self.keys()).union(other.keys())
        return fusioned_keys.difference(['datasets', 'count'])

    def _fusion_metadata_values(self, other, keys):
        return [(k, self._fusion_metadata_value(other, k))
                for k in keys]

    def to_json(self):
        data = self.convert_export_format()
        return json.dumps(data)

    def convert_export_format(self):
        datadict = {k: v
                    for k, v in self.items()
                    if k != 'datasets'}
        datadict['datasets'] = list(self['datasets'].values())
        return datadict

    def _fusion_metadata_value(self, other, key):
        a = self.get(key, 'NA')
        b = other.get(key, 'NA')
        return '{};{}'.format(a, b)

    @staticmethod
    def make_metadata(raw_metadata, unique_identifier):
        tuples = [(k, v)
                  for k, v in raw_metadata.items()
                  if k != 'datasets']
        tuples.append(('datasets',
                       _Datasets.make_datasets(
                           raw_metadata['datasets'], unique_identifier)))
        return _Metadata(tuples, unique_identifier)


class _Datasets(dict):
    def __init__(self, tuples, unique_identifier):
        super(_Datasets, self).__init__(tuples)
        self.unique_identifier = unique_identifier

    def __missing__(self, key):
        return _UserDataset([(self.unique_identifier, key)])

    def __add__(self, other):
        datasets = _Datasets(self, self.unique_identifier)
        datasets.update(other)
        return datasets

    def extract(self, file_names):
        return _Datasets([(k, self[k])
                          for k in file_names
                          if k in self],
                         self.unique_identifier)

    @staticmethod
    def make_datasets(raw_datasets, unique_identifier):
        return _Datasets([(k, _Dataset(v))
                          for k, v in raw_datasets.items()],
                         unique_identifier)


class _Dataset(dict):
    def __missing__(self, key):
        return 'NA'


class _UserDataset(_Dataset):
    def __missing__(self, key):
        return 'User'



from analysis.shared.clustering.scipypatch import ScipyPatch
from analysis.annotate.report.composers.composer import Composer, USER,\
    LegendMaker, PieChartsBuilder
from textwrap import TextWrapper
import pylab
import matplotlib.pyplot as plt
import warnings
import matplotlib
matplotlib.use('Agg')
warnings.filterwarnings("ignore")


class FirstPageComposer(Composer):
    def __init__(self, annotation, title, ordered_group_names,
                 rescale_heatmap):
        self.annotation = annotation
        self.title = title
        self.ordered_group_names = ordered_group_names
        self.rescale_heatmap = rescale_heatmap

    def run(self, left_lower_page, right_lower_page):
        ordered_matrix = self.annotation.get_matrix()

        if len(self.annotation) < 1785:
            scaling = 1785/float(len(self.annotation))
        else:
            scaling = 1

        matplotlib.rcParams['font.size'] = 14 * \
            scaling * len(self.annotation)/1785
        m = len(self.annotation) * scaling
        n = (14 * m) / 8.5

        x_size = int(n * 5/4.)
        y_size = int(m * 5/4.)

        fig = pylab.figure(figsize=(x_size/100., y_size/100.), dpi=100.)
        axdendro = fig.add_axes([0.0, 0.1, 0.1, 0.8])
        axdendro.axis('off')

        ScipyPatch.make_dendrogram(
            self.annotation.obtain_clusters(), self.annotation.get_matrix())
        axdendro.set_xticks([])
        axdendro.set_yticks([])

        # Plot distance matrix.
        axmatrix = fig.add_axes([0.11, 0.1, 0.4857142857, 0.8])
        axmatrix.axis('off')

        if self.rescale_heatmap:
            vmin, vmax = None, None
        else:
            vmin, vmax = -1, 1

        im = axmatrix.matshow(ordered_matrix.get_matrix(
        ), interpolation='nearest', cmap='RdBu_r', origin='upper', vmin=vmin,
            vmax=vmax)
        axmatrix.set_xticks([])
        axmatrix.set_yticks([])

        # User matrix line
        users_headers = self.annotation.obtain_file_names_with_tag_name(USER)
        users_line_pos = [-i
                          for i, h
                          in enumerate(ordered_matrix.get_file_names())
                          if h in users_headers]
        axuserline = fig.add_axes([0.5965, 0.1, 0.005, 0.8])
        axuserline.axis('off')

        plt.plot([1 for _ in range(len(users_line_pos))], users_line_pos, 'ro',
                 markersize=matplotlib.rcParams['font.size']/2.,
                 color=LegendMaker.USER_COLOR)
        pylab.xlim([0.5, 1.5])
        pylab.ylim([-len(ordered_matrix)+1, 0])
        axuserline.set_xticks([])
        axuserline.set_yticks([])

        # Plot colorbar.
        axcolor = fig.add_axes([0.11, 0.905, 0.4857142857, 0.025])
        pylab.colorbar(im, cax=axcolor, orientation='horizontal')
        axcolor.xaxis.set_ticks_position('top')
        pylab.title(self.title, fontsize=(
            matplotlib.rcParams['font.size'] * 4.2), y=2, loc='left')

        # Pie charts
        self._draw_aligned_piecharts(fig)

        # cluster id
        axcluster = fig.add_axes([0.102, 0.1, 0.005, 0.8])
        axcluster.axis('off')
        self._align_cluster_labels(axcluster, scaling)
        axcluster.set_xticks([])
        axcluster.set_yticks([])

        # description
        self._draw_description(fig)

        # time and page
        axid = fig.add_axes([0., 0., 1., 0.015])
        axid.axis('off')
        axid.text(0.005, 0.25, left_lower_page)
        axid.text(0.995, 0.25, right_lower_page, ha='right')
        axid.set_xticks([])
        axid.set_yticks([])

        # Display and save figure.
        filename = self.tmp_name()
        fig.savefig(filename, format='png')
        return filename

    def _draw_description(self, fig):
        formatted_description = '\n'.join(
            TextWrapper(
                width=140
            ).wrap(
                self.annotation.description))
        axdescription = fig.add_axes([0.11, 0.015, 0.4857142857, 0.075])
        axdescription.text(0, 1, formatted_description,
                           va='top', ha='left', clip_on=True, fontsize='large')
        axdescription.axis('off')
        axdescription.set_xticks([])
        axdescription.set_yticks([])

    def _compute_color_threshold(self, linkage_matrix, threshold):
        return linkage_matrix[-(threshold-1), 2]

    def _draw_aligned_piecharts(self, fig):
        for column, group in enumerate(self.ordered_group_names[:3]):
            piecharts_builder = PieChartsBuilder()
            piecharts_builder.set_annotation(self.annotation)
            piecharts_builder.set_category_name(group)
            piecharts_builder.set_max_nb_tags(12)
            piecharts_builder.set_max_chars(30)

            piecharts = piecharts_builder.build()

            self._draw_piecharts_lines(fig, piecharts, column)
            self._draw_legend(fig, piecharts, column)

    def _draw_legend(self, fig, piecharts, col_pos):
        if len(piecharts) <= 20:
            axlegend = fig.add_axes([0.60 + (col_pos * 0.1), 0.0, 0.1, 0.21])
            axlegend.axis('off')
            axlegend.set_xticks([])
            axlegend.set_yticks([])
            legend = plt.legend(handles=piecharts.get_handles(),
                                loc=2,
                                borderaxespad=0.,
                                title=piecharts.get_category_name())
            plt.setp(legend.get_title(), fontsize='xx-large')

    def _draw_piecharts_lines(self, fig, piecharts, column):
        nb_piecharts = len(piecharts)

        if nb_piecharts <= 10:
            line_a = reversed(range(nb_piecharts))
            line_b = range(0)
            stretch = 10 / float(nb_piecharts)
        elif nb_piecharts <= 20:
            side = nb_piecharts % 2  # 0 or 1
            line_a = reversed(range(1 - side, nb_piecharts, 2))
            line_b = reversed(range(0 + side, nb_piecharts, 2))
            stretch = 20 / float(nb_piecharts)

        if nb_piecharts <= 20:
            distance_between_charts = 0.07 * stretch
            first_axes = [0.60 + (column * 0.1), 0.215, 0.034, 0.056]
            second_axes = [0.64 + (column * 0.1), 0.235 +
                           0.020 * stretch, 0.034, 0.056]

            self._draw_piecharts_line(
                fig, piecharts, line_a, first_axes, distance_between_charts)
            self._draw_piecharts_line(
                fig, piecharts, line_b, second_axes, distance_between_charts)

    def _draw_piecharts_line(self, fig, piecharts, cluster_ids, axes, pos_inc):
        for pos, cluster_id in enumerate(cluster_ids):
            fig.add_axes(
                [axes[0], axes[1] + (pos * pos_inc), axes[2], axes[3]])
            pylab.pie(piecharts.get_percents_for_cluster(cluster_id),
                      colors=piecharts.get_colors_for_cluster(cluster_id),
                      shadow=True)
            pylab.title(piecharts.get_title_for_cluster(cluster_id), y=0.95)

    def _align_cluster_labels(self, ax, scaling):
        total = len(self.annotation)
        clusters = self.annotation.obtain_clusters()

        total_px = total * scaling

        # caculate pixels position for text and repositionning
        px_pos = self._compute_text_pos(clusters, total_px, scaling)

        # compile text
        for cluster_id in range(1, len(clusters) + 1):
            ax.text(0.0, px_pos[cluster_id] / float(total_px),
                    clusters[cluster_id].get_name(), fontsize='medium',
                    family='monospace')

    def _compute_text_pos(self, clusters, total_px, scaling):
        before = total_px
        pos_before = total_px
        px_pos = {}
        for cluster_id in range(1, len(clusters) + 1):
            px_pos[cluster_id] = before - \
                (len(clusters[cluster_id]) * scaling / 2)
            diff = (pos_before) - \
                (px_pos[cluster_id] + matplotlib.rcParams['font.size'] + 2)
            if (diff < 0):
                px_pos[cluster_id] += diff
            pos_before = px_pos[cluster_id]
            before -= len(clusters[cluster_id]) * scaling

        return px_pos



from .launcher import CatLaucher, GrepLauncher


class CatParser(object):
    @staticmethod
    def add_parser(subparsers):
        parser = subparsers.add_parser(
            'cat',
            description='UNIX cat like command for json metadata.',
            help='UNIX cat like command for json metadata.'
        )
        parser.set_defaults(func=CatLaucher.run)

        parser.add_argument(
            'input',
            nargs='+',
            help='The metadate file: metadata.json',
            type=str
        )
        parser.add_argument(
            '--md5sum',
            help="Set 'md5sum' as the unique identifier for files instead of \
                file_names.",
            action='store_true')


class GrepParser(object):
    @staticmethod
    def add_parser(subparsers):
        parser = subparsers.add_parser(
            'grep',
            description='UNIX grep like command for json metadata.',
            help='UNIX grep like command for json metadata.'
        )
        parser.set_defaults(func=GrepLauncher.run)

        parser.add_argument('-v', action='store_true',
                            help='Invert the match. Like not the pattern.')

        group = parser.add_mutually_exclusive_group()
        group.add_argument('-e', action='store_true',
                           help='The pattern is a regex for the research.')
        group.add_argument('-f', action='store_true',
                           help='The patterns are in a file.')

        parser.add_argument(
            'pattern', help='The pattern for the research.', type=str)
        parser.add_argument(
            'inputs', nargs='+', help='The raw metadate file: metadata.json',
            type=str)
        parser.add_argument(
            '--md5sum',
            help="Set 'md5sum' as the unique identifier for files instead of \
                file_names.",
            action='store_true')



import numpy
import warnings
from scipy.cluster.hierarchy import linkage, fcluster
from sklearn.metrics import silhouette_score

from .clusters import Clusters


class HierarchicalClustering(object):
    CLUSTERING_ALGORITHM = 'hierarchical'

    def __init__(self, matrix, linkage_method):
        self.matrix = matrix
        self.linkage_method = linkage_method
        self.fcluster_criterion = 'maxclust'

        if len(matrix) > 1:
            self.linkage_matrix = self._linkage(matrix, linkage_method)
        else:
            self.linkage_matrix = numpy.array([])

    def run(self, wanted_nb_clusters=None):
        if len(self.linkage_matrix) == 0:
            flat_cluster = [1]
        elif wanted_nb_clusters is None:
            flat_cluster = self._find_best_flat_clusters(self.linkage_matrix)
        else:
            flat_cluster = self._fcluster(
                self.linkage_matrix, wanted_nb_clusters,
                self.fcluster_criterion)

        return Clusters.make_clusters(
            flat_cluster, self.matrix.get_file_names(), self.linkage_matrix,
            self.linkage_method, self.CLUSTERING_ALGORITHM)

    @staticmethod
    def make_clusters(matrix, linkage_method, wanted_nb_clusters=None):
        return HierarchicalClustering(
            matrix, linkage_method).run(wanted_nb_clusters)

    def _find_best_flat_clusters(self, linkage_matrix):
        maximum = 21
        if maximum > len(self.matrix):
            maximum = len(self.matrix)

        best_nb_clusters = (-1.0, 2, range(len(self.matrix)))
        for i in range(2, maximum):
            flat_cluster = self._fcluster(
                linkage_matrix, i, self.fcluster_criterion)
            silhouette_avg = silhouette_score(
                self.matrix.get_matrix(), flat_cluster)
            if best_nb_clusters[0] < silhouette_avg:
                best_nb_clusters = (silhouette_avg, i, flat_cluster)

        return best_nb_clusters[2]

    def _linkage(self, matrix, linkage_method):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            return linkage(matrix.to_distance().get_matrix(), linkage_method)

    def _fcluster(self, linkage_matrix, asked_nb_clusters, fcluster_criterion):
        return fcluster(linkage_matrix, asked_nb_clusters, fcluster_criterion)



from analysis.manifold import MDSManifoldFactory, UMAPManifoldFactory
from .composers.mds_page_composer import MdsPageComposer
from .composers.pie_page_composer import PiePageComposer
from .composers.first_page_composer import FirstPageComposer
from time import localtime, strftime
import warnings
import matplotlib
matplotlib.use('Agg')
warnings.filterwarnings("ignore")


class Rendering:
    def __init__(self, tool_name):
        self.time = strftime("%a, %d %b %Y %H:%M:%S %Z", localtime())
        self.tool_name = tool_name

    def render(self, annotation, title, seed, mds_scatter_plot,
               umap_scatter_plot, ordered_group_names, rescale_heatmap=False):
        matrix = annotation.get_matrix()

        composers = [FirstPageComposer(
            annotation, title, ordered_group_names, rescale_heatmap)]

        for group_name in ordered_group_names:
            composers.append(PiePageComposer(annotation, title, group_name))

        if mds_scatter_plot:
            p_matrix = MDSManifoldFactory(seed).make().fit_transform(
                matrix.to_distance().get_matrix())
            if len(p_matrix) != len(matrix.get_file_names()):
                raise Exception("Ops! Something is went wrong with the MDS.")

            for group_name in ordered_group_names:
                composers.append(MdsPageComposer(
                    p_matrix, annotation, group_name, title, "MDS"))

        if umap_scatter_plot:
            p_matrix = UMAPManifoldFactory(seed).make().fit_transform(
                matrix.to_distance().get_matrix())
            if len(p_matrix) != len(matrix.get_file_names()):
                raise Exception("Ops! Something is went wrong with the UMAP.")

            for group_name in ordered_group_names:
                composers.append(MdsPageComposer(
                    p_matrix, annotation, group_name, title, "UMAP"))

        return [composers[i].run(self.make_left_lower_page(),
                                 self.make_rigth_lower_page(i + 1))
                for i in range(len(composers))]

    def make_left_lower_page(self):
        return self.time + '; ' + self.tool_name

    def make_rigth_lower_page(self, page_number):
        return 'Page ' + str(page_number)



from setuptools import setup, find_packages
from analysis.config import VERSION, DESCRIPTION, NAME

with open('README.md', 'r') as f:
    LONG_DESCRIPTION = f.read()

setup(
    name=NAME,
    version=VERSION,
    author="Simon Hébert-Deschamps",
    author_email="simon.hebert-deschamps@usherbrooke.ca",
    description=DESCRIPTION,
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'epigeec-analysis = analysis.main:cli',
        ]
    },
    test_suite='tests',
    install_requires=[
        'matplotlib',
        'numpy',
        'pandas',
        'scipy',
        'sklearn',
        'reportlab',
        'umap-learn',
    ],
    python_requires='>=3.6.*',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    url='https://bitbucket.org/labjacquespe/epigeec_analysis',
    classifiers=[
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Intended Audience :: Science/Research',
        'Environment :: Console',
    ],
)



from .launcher import EvaluateLauncher


class EvaluateParser(object):
    @staticmethod
    def add_parser(subparsers):
        parser = subparsers.add_parser(
            'evaluate',
            description='Evaluate the matrix',
            help='Evaluate the matrix')
        parser.set_defaults(func=EvaluateLauncher.run)

        parser.add_argument(
            'matrix', help='The input matrix tsv file.', type=str)
        parser.add_argument(
            'metadata', help='The input metadata json file.', type=str)
        parser.add_argument(
            '--linkage',
            help='The linkage method. By default: "average" Options: "average"\
                , "complete", "weighted" or "single".',
            default='average',
            choices=['average', 'complete', 'weighted', 'single'],
            type=str)
        parser.add_argument(
            '--title',
            help='The title of the report.',
            default='epiGeEC Ari 1.0.0',
            type=str)
        parser.add_argument(
            '-b',
            nargs='*',
            help='The breakdown categories. If not specified, it will use all \
                categories in metadata.',
            default=[],
            metavar='category',
            type=str)
        parser.add_argument(
            '-r',
            nargs='*',
            help='The annotation categories restrictions. If not specified, it \
                will use all categories in metadata.',
            default=[],
            metavar='category',
            type=str)
        parser.add_argument(
            '-i',
            help='ignore User',
            action='store_true')
        parser.add_argument(
            '--md5sum',
            help="Set 'md5sum' as the unique identifier for files instead of \
                file_names.",
            action='store_true')
